/* Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include <udm_config.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_ARPA_NAMESER_H
#include <arpa/nameser.h>
#endif
#ifdef HAVE_RESOLV_H
#include <resolv.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif

#include "udm_store.h"
#include "udm_services.h"
#include "udm_xmalloc.h"
#include "udm_crc32.h"
#include "udm_utils.h"
#include "udm_log.h"
#include "udm_vars.h"

#ifdef HAVE_ZLIB


#ifndef HAVE_HSTRERROR
static const char *h_errlist[] = {
    "Error 0",
    "Unknown host",
    "Host name lookup failure",
    "Unknown server error",
    "No address associated with name"
};
# define hstrerror(err)  ((err) <= 4 ? h_errlist[(err)] : "unknown error")
#endif


#endif


__INDLIB__ int UdmStoreDoc(UDM_AGENT *Agent, UDM_DOCUMENT *Doc) {

#ifdef HAVE_ZLIB
  const char *hello = "S\0";
  char result[8];
  int s;
  size_t content_size=Doc->Buf.size - (Doc->Buf.content-Doc->Buf.buf);
  const char *url = UdmVarListFindStr(&Doc->Sections, "URL", "");
  int rec_id=UdmCRC32(url, strlen(url));
  
  if ((s = UdmStoreOpen(Agent)) < 0)return -1;
  
  /* FIXME: add checking of send() results */
  UdmSend(s, hello, 1, 0);
  UdmSend(s, &rec_id, sizeof(rec_id), 0);
  UdmSend(s, &content_size, sizeof(content_size), 0);
  UdmSend(s, Doc->Buf.content, content_size, 0);
  
  if (UdmRecvall(s, result, 6) < 0) {
    closesocket(s);
    return -1;
  }
  closesocket(s);
  
  return memcmp(result, "STORED", 6) ? -1 : 0;
  
#else
  return -1;
#endif

}


/* Retrieve cached copy       */
/* Caller must alloc Doc->Buf.buf */

__INDLIB__ int UdmUnStoreDoc(UDM_AGENT *Agent, UDM_DOCUMENT *Doc) {
  
#ifdef HAVE_ZLIB
  const char *hello = "G\0";
  int s;
  unsigned int rec_id;
  size_t content_size=0;
  ssize_t nread;
  
  sscanf(UdmVarListFindStr(&Doc->Sections, "STORED_ID", "0"), "%u", &rec_id); /*!!! Do not replace this by UdmVarListFindInt */
  Doc->Buf.size=0;
  
  if ((s = UdmStoreOpen(Agent)) < 0) return -1;
  
  /* FIXME: add send() results checking */
  UdmSend(s, hello, 1, 0);
  UdmSend(s, &rec_id, sizeof(rec_id), 0);
  
  if (
       (UdmRecvall(s, &content_size, sizeof(content_size)) < 0) ||
       (content_size == 0)     ) 
  {
    closesocket(s);
    return -1;
  }
 
  if (Doc->Buf.buf == NULL) {
    Doc->Buf.buf = (char*) malloc(content_size + 1);
  }
  if (
      (Doc->Buf.buf == NULL) ||
       ((nread = UdmRecvall(s, Doc->Buf.buf, content_size)) < 0)
      ) {
    closesocket(s);
    return -1;
  }
  
  closesocket(s);
  Doc->Buf.buf[nread] = '\0';
  Doc->Buf.size = nread;
  Doc->Buf.content=Doc->Buf.buf;
  return 0;
  
#else
  return -1;
#endif
}

int UdmStoreOpen(UDM_AGENT *Agent) {

#ifdef HAVE_ZLIB
  struct hostent *hp;
  struct sockaddr_in server_addr;
  int s, nport = UDM_STORED_PORT;
  char * cport;
  char * stored_addr;
  const char *sa = UdmVarListFindStr(&Agent->Conf->Vars, "StoredAddr", NULL);

  if(!sa) return -1;
  stored_addr = strdup(sa);

  if((cport=strchr(stored_addr,':'))){
    *cport='\0';
    nport=atoi(cport+1);
  }

  if ((hp = gethostbyname(stored_addr)) == 0 ) {
    free(stored_addr);
    UdmLog(Agent, UDM_LOG_ERROR, "StoreOpen ERR gethostbyname: %s", hstrerror(h_errno));
    return -1;
  }
  free(stored_addr);
  bzero(&server_addr, sizeof(server_addr));
  memmove(&server_addr.sin_addr, hp->h_addr, (size_t)hp->h_length);
  server_addr.sin_family = hp->h_addrtype;
  server_addr.sin_port = htons((u_short)nport);

  if((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    UdmLog(Agent, UDM_LOG_ERROR, "StoreOpen ERR socket: %s", strerror(errno));
    return -1;
  }
  
  if(connect(s, (struct sockaddr *)&server_addr, sizeof(server_addr)) == -1) {
    UdmLog(Agent, UDM_LOG_ERROR, "StoreOpen ERR connect to %s: %s", inet_ntoa(server_addr.sin_addr), strerror(errno));
    closesocket(s);
    return -1;
  }

  return s;

#else
  return -1;
#endif

}


__INDLIB__ int UdmStoreDeleteDoc(UDM_AGENT *Agent, UDM_DOCUMENT *Doc) {

#ifdef HAVE_ZLIB
  const char *hello = "D\0";
  int s;
  const char *url = UdmVarListFindStr(&Doc->Sections, "URL", "");
  int rec_id=UdmCRC32(url, strlen(url));

  if ((s = UdmStoreOpen(Agent)) < 0) return -1;
  
  UdmSend(s, hello, 1, 0);
  UdmSend(s, &rec_id, sizeof(rec_id), 0);
  closesocket(s);
  return 0;

#else
  return -1;
#endif

}
